@@ -272,16 +272,18 @@ static int iswnumeric( const wchar_t *n )
 */
 static int match_pid( const wchar_t *cmd,
 					  const wchar_t *proc,
-					  int flags )
+					  int flags,
+					  int *offset)
 {
 	/* Test for direct match */
-
+	
 	if( wcsncmp( cmd, proc, wcslen( proc ) ) == 0 )
+	{
+		if( offset )
+			*offset = 0;
 		return 1;
-
-	if( flags & ACCEPT_INCOMPLETE )
-		return 0;
-
+	}
+	
 	/*
 	  Test if the commandline is a path to the command, if so we try
 	  to match against only the command part
@@ -309,7 +311,11 @@ static int match_pid( const wchar_t *cmd,
 
 		if( wcsncmp( start+1, proc, wcslen( proc ) ) == 0 )
 		{
+			if( offset )
+				*offset = start+1-first_token;
+			
 			free( first_token );
+
 			return 1;
 		}
 	}
@@ -381,8 +387,12 @@ static int find_process( const wchar_t *proc,
 		else
 		{
 
-			int jid = wcstol( proc, 0, 10 );
-			if( jid > 0 )
+			int jid;
+			wchar_t *end;
+			
+			errno = 0;
+			jid = wcstol( proc, &end, 10 );
+			if( jid > 0 && !errno && !*end )
 			{
 				j = job_get( jid );
 				if( (j != 0) && (j->command != 0 ) )
@@ -403,14 +413,16 @@ static int find_process( const wchar_t *proc,
 
 	for( j=first_job; j != 0; j=j->next )
 	{
+		int offset;
+		
 		if( j->command == 0 )
 			continue;
-
-		if( match_pid( j->command, proc, flags ) )
+		
+		if( match_pid( j->command, proc, flags, &offset ) )
 		{
 			if( flags & ACCEPT_INCOMPLETE )
 			{
-				wchar_t *res = wcsdupcat2( j->command + wcslen(proc),
+				wchar_t *res = wcsdupcat2( j->command + offset + wcslen(proc),
 										   COMPLETE_SEP_STR,
 										   COMPLETE_JOB_DESC,
 										   (void *)0 );
@@ -438,15 +450,16 @@ static int find_process( const wchar_t *proc,
 			continue;
 		for( p=j->first_process; p; p=p->next )
 		{
-
+			int offset;
+			
 			if( p->actual_cmd == 0 )
 				continue;
 
-			if( match_pid( p->actual_cmd, proc, flags ) )
+			if( match_pid( p->actual_cmd, proc, flags, &offset ) )
 			{
 				if( flags & ACCEPT_INCOMPLETE )
 				{
-					wchar_t *res = wcsdupcat2( p->actual_cmd + wcslen(proc),
+					wchar_t *res = wcsdupcat2( p->actual_cmd + offset + wcslen(proc),
 											   COMPLETE_SEP_STR,
 											   COMPLETE_CHILD_PROCESS_DESC,
 											   (void *)0);
@@ -556,16 +569,19 @@ static int find_process( const wchar_t *proc,
 
 		if( cmd != 0 )
 		{
-			if( match_pid( cmd, proc, flags ) )
+			int offset;
+			
+			if( match_pid( cmd, proc, flags, &offset ) )
 			{
 				if( flags & ACCEPT_INCOMPLETE )
 				{
-					wchar_t *res = wcsdupcat2( cmd + wcslen(proc),
+					wchar_t *res = wcsdupcat2( cmd + offset + wcslen(proc),
 											   COMPLETE_SEP_STR,
 											   COMPLETE_PROCESS_DESC,
 											   (void *)0);
 					if( res )
 						al_push( out, res );
+
 				}
 				else
 				{
@@ -1568,8 +1584,14 @@ int expand_string( void *context,
 		{
 			wchar_t *next;
 
-			next = expand_unescape( (wchar_t *)al_get( in, i ),
-									1);
+			/*
+			  We accept incomplete strings here, since complete uses
+			  expand_string to expand incomplete strings from the
+			  commandline.
+			*/
+			int unescape_flags = UNESCAPE_SPECIAL | UNESCAPE_INCOMPLETE;
+
+			next = expand_unescape( (wchar_t *)al_get( in, i ), unescape_flags );
 
 			free( (void *)al_get( in, i ) );
 
@@ -1731,6 +1753,15 @@ int expand_string( void *context,
 						al_truncate( out, 0 );
 						break;
 					}
+
+					case -1:
+					{
+						al_foreach( out, &free );
+						al_destroy( in );
+						al_destroy( out );
+						return EXPAND_ERROR;
+					}
+					
 				}
 			}
 			else
