@@ -48,7 +48,8 @@ enum
 	R_VI_DELETE_TO,
 	R_EXECUTE,
 	R_BEGINNING_OF_BUFFER,
-	R_END_OF_BUFFER
+	R_END_OF_BUFFER,
+	R_REPAINT
 }
 ;
 
