@@ -15,6 +15,17 @@
 #define PARSER_TEST_ERROR 1
 #define PARSER_TEST_INCOMPLETE 2
 
+/**
+   REturn valuse for parser_is_switch()
+*/
+enum 
+{
+	ARG_NON_SWITCH,
+	ARG_SWITCH,
+	ARG_SKIP
+}
+	;
+
 /**
    event_block_t represents a block on events of the specified type
 */
@@ -220,6 +231,13 @@ int eval_args( const wchar_t *line,
 void error( int ec, int p, const wchar_t *str, ... );
 
 
+/**
+   Check if the specified argument is a switch. Return ARG_SWITCH if yes,
+   ARG_NON_SWITCH if no and ARG_SKIP if the argument is '--'
+*/
+int parser_is_switch( const wchar_t *cmd );
+
+
 /**
    Tests if the specified commands parameters should be interpreted as another command, which will be true if the command is either 'command', 'exec', 'if', 'while' or 'builtin'.  
 
