@@ -28,22 +28,17 @@ segments.
 /**
    Error string for unexpected end of string
 */
-#define EOL_ERROR _( L"Unexpected end of token" )
-
-/**
-   Error string for unexpected end of string
-*/
-#define QUOTE_ERROR _( L"Unterminated quote" )
+#define QUOTE_ERROR _( L"Unexpected end of string, quotes are not balanced" )
 
 /**
    Error string for mismatched parenthesis
 */
-#define PARAN_ERROR _( L"Parenthesis mismatch" )
+#define PARAN_ERROR _( L"Unexpected end of string, parenthesis do not match" )
 
 /**
    Error string for invalid redirections
 */
-#define REDIRECT_ERROR _( L"Invalid redirection" )
+#define REDIRECT_ERROR _( L"Invalid input/output redirection" )
 
 /**
    Error string for when trying to pipe from fd 0
@@ -142,27 +137,7 @@ void tok_init( tokenizer *tok, const wchar_t *b, int flags )
 
 	tok->has_next = (*b != L'\0');
 	tok->orig_buff = tok->buff = (wchar_t *)(b);
-
-	if( tok->accept_unfinished )
-	{
-		int l = wcslen( tok->orig_buff );
-		if( l != 0 )
-		{
-			if( tok->orig_buff[l-1] == L'\\' )
-			{
-				tok->free_orig = 1;
-				tok->orig_buff = tok->buff = wcsdup( tok->orig_buff );
-				if( !tok->orig_buff )
-				{
-					DIE_MEM();
-				}
-				tok->orig_buff[l-1] = L'\0';
-			}
-		}
-	}
-
 	tok_next( tok );
-
 }
 
 void tok_destroy( tokenizer *tok )
@@ -207,11 +182,7 @@ int tok_has_next( tokenizer *tok )
 
 static int is_string_char( wchar_t c )
 {
-	if( !c ||  wcschr( SEP, c ) )
-	{
-		return 0;
-	}
-	return 1;
+	return !( !c ||  wcschr( SEP, c ) );
 }
 
 /**
@@ -251,8 +222,17 @@ static void read_string( tokenizer *tok )
 				tok->buff++;
 				if( *tok->buff == L'\0' )
 				{
-					tok_error( tok, TOK_UNTERMINATED_ESCAPE, EOL_ERROR );
-					return;
+					if( (!tok->accept_unfinished) )
+					{
+						tok_error( tok, TOK_UNTERMINATED_ESCAPE, QUOTE_ERROR );
+						return;
+					}
+					else
+					{
+						do_loop = 0;
+					}
+					
+
 				}
 				else if( *tok->buff == L'\n' && mode == 0)
 				{
@@ -683,7 +663,7 @@ int tok_get_pos( tokenizer *tok )
 {
 	CHECK( tok, 0 );
 	
-	return tok->last_pos + (tok->free_orig?1:0);
+	return tok->last_pos;
 }
 
 
